# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import LogLayer

from .base import BaseTransformer


class LogParams(Params):
    """
    Mixin class containing alpha parameter needed for log transform layers.
    """

    alpha = Param(
        Params._dummy(),
        "alpha",
        "Value to use in log transform: log(alpha + x)",
        typeConverter=TypeConverters.toFloat,
    )

    def setAlpha(self, value: float) -> "LogParams":
        """
        Sets the alpha parameter.

        :param value: Float value to use in log transform: log(alpha + x).
        :returns: Instance of class mixed in.
        """
        return self._set(alpha=value)

    def getAlpha(self) -> float:
        """
        Gets the alpha parameter.

        :returns: Float value of alpha used in log transform.
        """
        return self.getOrDefault(self.alpha)


class LogTransformer(
    BaseTransformer,
    LogParams,
    SingleInputSingleOutputParams,
):
    """
    Log Spark Transformer for use in Spark pipelines.
    This transformer applies a log(alpha + x) transform to the input column.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        alpha: float = 0.0,
    ) -> None:
        """
        Instantiates a LogTransformer transformer. Sets the default values of:

        - alpha: 0

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param alpha: Value to use in log transform: log(alpha + x). Default is 0.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(alpha=0.0)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column named outputCol with the
        log transform of the inputCol.

        :param dataset: Pyspark DataFrame to transform.
        :returns: Transformed pyspark dataFrame.
        """
        alpha = self.getAlpha()
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: F.log(x + F.lit(alpha)),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the log transform.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the log(alpha + x) operation.
        """
        return LogLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            alpha=self.getAlpha(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ImputeLayer(BaseLayer):
    """
    Performs imputation on the input.
    Where the input data is equal to the specified mask value, this layer will replace
    the data with the impute value calculated at preprocessing time.
    The impute value is either the mean or median and is computed while ignoring rows
    in the data which are equal to the mask value or are null.
    """

    def __init__(
        self,
        impute_value: Union[float, str, int],
        mask_value: Union[float, str, int],
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialise the ImputeLayer layer.
        :param impute_value: The value to use for imputation.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param mask_value: Value which should be replaced by the
        impute value at inference.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.impute_value = impute_value
        self.mask_value = mask_value
        if not isinstance(self.mask_value, type(self.impute_value)):
            raise ValueError(
                "The mask value and impute value must be of the same type."
            )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs imputation on the input tensor(s) by calling the keras
        ImputeLayer layer. It imputes over values which are equal to the
        mask_value.
        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.
        :param inputs: Input tensor to perform the imputation on.
        :returns: The input tensor with the imputation applied.
        """
        if inputs.dtype.is_floating or inputs.dtype.is_integer:
            inputs, mask = self._force_cast_to_compatible_numeric_type(
                inputs, self.mask_value
            )
            inputs, impute_value = self._force_cast_to_compatible_numeric_type(
                inputs, self.impute_value
            )
        else:
            mask = self._cast(tf.constant(self.mask_value), inputs.dtype.name)
            impute_value = self._cast(tf.constant(self.impute_value), inputs.dtype.name)

        mask = tf.equal(inputs, mask)
        imputed_outputs = tf.where(
            mask,
            impute_value,
            inputs,
        )

        return imputed_outputs

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the ImputeLayer layer.
        Used for saving and loading from a model.
        Specifically adds additional parameters to the base configuration.
        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "impute_value": self.impute_value,
                "mask_value": self.mask_value,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input
from kamae.utils import get_condition_operator

from .base import BaseLayer


# TODO: Deprecate this in favor of IfStatementLayer in next major release.
@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class NumericalIfStatementLayer(BaseLayer):
    """
    Performs a numerical if statement on the input tensor,
    returning a tensor of the same shape as the input tensor.

    The condition operator can be one of the following:
    - "eq": Equal to
    - "neq": Not equal to
    - "lt": Less than
    - "le": Less than or equal to
    - "gt": Greater than
    - "ge": Greater than or equal to

    The value to compare must be a float. We will cast the input tensor to a float
    if it is not already a float.

    If the condition is true, the result is the result_if_true value.
    If the condition is false, the result is the result_if_false value.

    If any of [value_to_compare, result_if_true, result_if_false] are None, we assume
    they are passed in as inputs to the layer in the above order. If all of them are
    not None, then inputs is expected to be a tensor.
    """

    def __init__(
        self,
        condition_operator: str,
        value_to_compare: float = None,
        result_if_true: float = None,
        result_if_false: float = None,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the NumericalIfStatementLayer layer.

        :param condition_operator: Operator to use in the if statement. Can be one of:
            - "eq": Equal to
            - "neq": Not equal to
            - "lt": Less than
            - "leq": Less than or equal to
            - "gt": Greater than
            - "geq": Greater than or equal to
        :param value_to_compare: Float value to compare the input tensor to. If None, we
        assume it is passed in as an input to the layer.
        :param result_if_true: Float value to return if the condition is true. If None,
        we assume it is passed in as an input to the layer.
        :param result_if_false: Float value to return if the condition is false. If
        None, we assume it is passed in as an input to the layer.
        :param name: The name of the layer. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.condition_operator = condition_operator
        self.value_to_compare = value_to_compare
        self.result_if_true = result_if_true
        self.result_if_false = result_if_false

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bfloat16, tf.float16, tf.float32, tf.float64]

    def _construct_input_tensors(
        self, inputs: Iterable[tf.Tensor]
    ) -> Iterable[tf.Tensor]:
        """
        Constructs the input tensors for the layer in the case where all the optional
        parameters are not specified. We need to run through the provided inputs and
        either select an input or the specified parameter.

        Specifically for this layer, we assume the inputs are in the following order:
        [input_tensor, value_to_compare, result_if_true, result_if_false]

        Any but the input tensor can be None.

        :param inputs: List of input tensors.
        :returns: List of input tensors potentially containing constant tensors for the
        optional parameters.
        """
        optional_params = [
            self.value_to_compare,
            self.result_if_true,
            self.result_if_false,
        ]
        # Setup the inputs. Keep a counter to know how many tensors from inputs have
        # been used.
        input_col_counter = 1
        # First input is always the input tensor
        multiple_inputs = [inputs[0]]
        for param in optional_params:
            if param is None:
                # If the param is None, we assume it is an input tensor at the next
                # index
                multiple_inputs.append(inputs[input_col_counter])
                input_col_counter += 1
            else:
                # Otherwise, we create a constant tensor for the parameter
                # and do not increment the counter.
                multiple_inputs.append(tf.constant(param, dtype=inputs[0].dtype))
        return multiple_inputs

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs) -> Tensor:
        """
        Performs the numerical if statement on the inputs. If the inputs are a tensor,
        we assume that the value_to_compare, result_if_true, and result_if_false are
        provided. If the inputs are not a tensor, we assume any not provided are
        provided as inputs to the layer.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param inputs: Tensor or list of tensors.
        :returns: Tensor after computing the numerical if statement.
        """
        condition_op = get_condition_operator(self.condition_operator)
        if not len(inputs) > 1:
            # If the input is a tensor, we assume that the value_to_compare,
            # result_if_true, and result_if_false are provided
            if any(
                [
                    v is None
                    for v in [
                        self.value_to_compare,
                        self.result_if_true,
                        self.result_if_false,
                    ]
                ]
            ):
                raise ValueError(
                    "If inputs is a tensor, value_to_compare, result_if_true, and "
                    "result_if_false must be specified."
                )
            cond = tf.where(
                condition_op(inputs[0], self.value_to_compare),
                tf.constant(self.result_if_true, dtype=inputs[0].dtype),
                tf.constant(self.result_if_false, dtype=inputs[0].dtype),
            )
            return cond
        else:
            # If the input is a list, we assume that the value_to_compare,
            # result_if_true, and result_if_false are potentially provided in the inputs
            input_tensors = self._construct_input_tensors(inputs)
            cond = tf.where(
                condition_op(input_tensors[0], input_tensors[1]),
                input_tensors[2],
                input_tensors[3],
            )
            return cond

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the NumericalIfStatement layer.

        Specifically adds the following to the base configuration:
        - condition_operator
        - value_to_compare
        - result_if_true
        - result_if_false

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "condition_operator": self.condition_operator,
                "value_to_compare": self.value_to_compare,
                "result_if_true": self.result_if_true,
                "result_if_false": self.result_if_false,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringContainsLayer(BaseLayer):
    """
    Performs a string contains operation on the input tensor,
    matching against a string constant or element-wise against a second input tensor.
    WARNING: While it works, the use of tensors in matching/replacement
        is not recommended due to the complexity of the regex matching which requires
        use of a map_fn. This will be comparatively VERY slow and may not be suitable
        for inference use-cases.
        If you know where in the string the match is, you will be much
        better off slicing the string and checking for equality.
    This implementation will only match an empty string with another empty string and
    does not support matching of newline characters.
    """

    def __init__(
        self,
        string_constant: Optional[str] = None,
        negation: bool = False,
        name: Optional[str] = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the StringContainsLayer layer.
        :param string_constant: The string to match against. Defaults to `None`.
        :param negation: Whether to negate the output. Defaults to `False`.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.negation = negation
        self.string_constant = string_constant

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs) -> Tensor:
        """
        Checks for the existence of a substring/pattern within a tensor.
        WARNING: While it works, the use of tensors in matching
        is not recommended due to the complexity of the regex matching which requires
        use of a map_fn. This will be comparatively VERY slow and may not be suitable
        for inference use-cases.
        If you know where in the string the match is, you will be much
        better off slicing the string and checking for equality.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param: inputs: A string tensor or iterable of up to two string tensors.
            In the case two tensors are passed, require that the first tensor is the
            tensor to match a pattern/substring against.
        :returns: A boolean tensor whether the string/string elements are matched.
        """

        match_all_pattern = ".*"

        # Checking input
        if self.string_constant is not None:
            if len(inputs) == 1:
                # To preserve shape, need to pass tensor to regex_full_match
                input_tensor = inputs[0]

                match_substring = self.string_constant
                match_substring = self._escape_special_characters(match_substring)
                matched_tensor = tf.strings.regex_full_match(
                    input_tensor,
                    tf.constant(
                        match_all_pattern + match_substring + match_all_pattern
                        if match_substring != ""
                        else "^$"
                    ),
                )
            else:
                raise ValueError(
                    "With string_constant defined, expected a single tensor as input."
                )
        else:
            if len(inputs) != 2:
                raise ValueError(
                    "Expected iterable of tensors of length 2, \
                     or string_constant to be defined."
                )

            # Two tensors provided
            @tf.function
            def tensor_match(x):
                match_substring = x[1]
                match_substring = self._escape_special_characters(match_substring)
                return tf.strings.regex_full_match(
                    x[0],
                    match_all_pattern + match_substring + match_all_pattern
                    if x[1] != ""
                    else "^$",
                )

            # Stack inputs to match element-wise with map_fn
            # Requires ordering of inputs to be correct
            stacked_inputs = tf.stack(inputs, axis=-1)
            input_shape = tf.shape(inputs[0])

            mappable_tensor = tf.reshape(stacked_inputs, [-1, 2])

            # Apply element-wise matching
            # TODO: tf.vectorized_map may be slightly faster with larger batches
            #  but this requires some refactoring
            matched_tensor = tf.map_fn(
                fn=tensor_match, elems=mappable_tensor, dtype=tf.bool
            )

            matched_tensor = tf.reshape(matched_tensor, input_shape)

        output_tensor = (
            tf.math.logical_not(matched_tensor) if self.negation else matched_tensor
        )

        return output_tensor

    def _escape_special_characters(
        self, string: Union[str, Tensor]
    ) -> Union[str, Tensor]:
        """
        Escapes special characters in a string so they are not parsed as regex.
        :param string: The string or string tensor to escape special characters in.
        :returns: The escaped string or string tensor.
        """
        escaped_string = string
        for char in [
            "\\",
            ".",
            "^",
            "$",
            "*",
            "+",
            "?",
            "{",
            "}",
            "[",
            "]",
            "(",
            ")",
            "|",
        ]:
            if isinstance(escaped_string, str):
                escaped_string = escaped_string.replace(char, "\\" + char)
            else:
                escaped_string = tf.strings.regex_replace(
                    escaped_string, "\\" + char, "\\" + char
                )
        return escaped_string

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringContains layer.
        Used for saving and loading from a model.

        Specifically adds the string_constant and negation parameters to the config
        dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {"string_constant": self.string_constant, "negation": self.negation}
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import InputOutputExtractor  # noqa: F401 # isort:skip
from .name import HasLayerName  # noqa: F401 # isort:skip
from .base import (  # noqa: F401 # isort:skip
    MultiInputMultiOutputParams,
    MultiInputSingleOutputParams,
    SingleInputMultiOutputParams,
    SingleInputSingleOutputParams,
    HasInputDtype,
    HasOutputDtype,
)
from .shared import (  # noqa: F401
    AutoBroadcastParams,
    ConstantStringArrayParams,
    DateTimeParams,
    DefaultIntValueParams,
    DropUnseenParams,
    HashIndexParams,
    ImputeMethodParams,
    LabelsArrayParams,
    LatLonConstantParams,
    ListwiseStatisticsParams,
    MaskValueParams,
    MathFloatConstantParams,
    NanFillValueParams,
    NegationParams,
    PadValueParams,
    StandardScaleParams,
    StandardScaleSkipZerosParams,
    StringConstantParams,
    StringIndexParams,
    StringRegexParams,
    UnixTimestampParams,
)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import (
    datetime_to_unix_timestamp,
    enforce_single_tensor_input,
)

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class DateTimeToUnixTimestampLayer(BaseLayer):
    """
    Returns the unix timestamp from a datetime in either yyyy-MM-dd HH:mm:ss.SSS
    or yyyy-MM-dd format.
    """

    def __init__(
        self,
        name=None,
        input_dtype: str = None,
        output_dtype: str = None,
        unit: str = "s",
        **kwargs,
    ) -> None:
        """
        Initialises an instance of the DateTimeToUnixTimstamp layer.

        :param name: Name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param unit: Unit of the timestamp. Can be `milliseconds` (or `ms`)
        or `seconds` (or `s`). Defaults to `s`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if unit not in ["milliseconds", "seconds", "ms", "s"]:
            raise ValueError(
                """Unit must be one of ["milliseconds", "seconds", "ms", "s"]"""
            )
        if unit == "milliseconds":
            unit = "ms"
        if unit == "seconds":
            unit = "s"
        self.unit = unit

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> tf.Tensor:
        """
        Returns the unix timestamp from a datetime in either yyyy-MM-dd HH:mm:ss.SSS
        or yyyy-MM-dd format.

        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to determine the shape of the output tensor.
        :returns: Unix timestamp in either milliseconds or seconds.
        """
        # Timestamp needs to be in float64 for unix_timestamp_to_datetime
        unix_timestamp_in_seconds = datetime_to_unix_timestamp(inputs)
        return (
            unix_timestamp_in_seconds
            if self.unit == "s"
            else unix_timestamp_in_seconds * 1000.0
        )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the DateTimeToUnixTimstamp layer.
        Used for saving and loading from a model.

        Specifically sets the `unit` parameters in the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "unit": self.unit,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Dict

import tensorflow as tf
from pyspark.sql.types import (
    BooleanType,
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
)


class DType(Enum):
    """
    Enum class for supported data types in Kamae.
    Contains a string name, the corresponding Spark data type, the corresponding
    TensorFlow data type, and the number of bytes the data type takes up.
    String is a special case, as it can be of any length, so the number of bytes
    is set to 0.
    """

    STRING = (
        "string",
        StringType(),
        tf.string,
        0,
        False,
        False,
    )  # String can be of any length
    BIGINT = ("bigint", LongType(), tf.int64, 8, False, True)
    INT = ("int", IntegerType(), tf.int32, 4, False, True)
    SMALLINT = ("smallint", ShortType(), tf.int16, 2, False, True)
    TINYINT = ("tinyint", ByteType(), tf.int8, 1, False, True)
    FLOAT = ("float", FloatType(), tf.float32, 4, True, False)
    DOUBLE = ("double", DoubleType(), tf.float64, 8, True, False)
    BOOLEAN = ("boolean", BooleanType(), tf.bool, 1, False, False)

    def __init__(
        self,
        dtype_name: str,
        spark_dtype: DataType,
        tf_dtype: tf.dtypes.DType,
        bytes: int,
        is_floating: bool = False,
        is_integer: bool = False,
    ) -> None:
        self.dtype_name = dtype_name
        self.spark_dtype = spark_dtype
        self.tf_dtype = tf_dtype
        self.bytes = bytes
        self.is_floating = is_floating
        self.is_integer = is_integer

    def as_dict(self) -> Dict[str, Any]:
        return {
            "dtype_name": self.dtype_name,
            "spark_dtype": self.spark_dtype,
            "tf_dtype": self.tf_dtype,
            "bytes": self.bytes,
            "is_floating": self.is_floating,
            "is_integer": self.is_integer,
        }



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import numpy as np
import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import SingleInputSingleOutputParams, StandardScaleParams
from kamae.spark.utils import single_input_single_output_array_transform
from kamae.tensorflow.layers import StandardScaleLayer

from .base import BaseTransformer


class StandardScaleTransformer(
    BaseTransformer,
    StandardScaleParams,
    SingleInputSingleOutputParams,
):
    """
    Standard scaler transformer for use in Spark pipelines.
    This is used to standardize/transform the input column
    using the mean and standard deviation.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        mean: Optional[List[float]] = None,
        stddev: Optional[List[float]] = None,
        maskValue: Optional[float] = None,
    ) -> None:
        """
        Initializes a StandardScaleTransformer transformer.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model.
        :param mean: List of mean values corresponding to the input column.
        :param stddev: List of standard deviation values corresponding to the
        input column.
        :param maskValue: Value which should be ignored in the standard scaling process.
        That is disregarded when calculating the mean and standard deviation and
        when the scaling is applied.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(maskValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset using the mean and standard deviation
        to standardize the input column. If a mask value is set, it is used
        to ignore elements in the dataset with that value, and they will remain
        unchanged in the standardization process.

        :param dataset: Pyspark dataframe to transform.
        :returns: Pyspark dataframe with the input column standardized,
         named as the output column.
        """
        original_input_datatype = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(original_input_datatype, ArrayType):
            input_col = F.array(F.col(self.getInputCol()))
            input_datatype = ArrayType(original_input_datatype)
        else:
            input_col = F.col(self.getInputCol())
            input_datatype = original_input_datatype

        shift = F.array([F.lit(m) for m in self.getMean()])
        scale = F.array([F.lit(1.0 / s if s != 0 else 0.0) for s in self.getStddev()])

        output_col = single_input_single_output_array_transform(
            input_col=input_col,
            input_col_datatype=input_datatype,
            func=lambda x: F.transform(
                x,
                lambda y, i: F.when(y == self.getMaskValue(), y).otherwise(
                    (y - F.lit(shift)[i]) * F.lit(scale)[i]
                ),
            ),
        )

        if not isinstance(original_input_datatype, ArrayType):
            output_col = output_col.getItem(0)

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the standard scaler transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
         that performs the standardization.
        """
        np_mean = np.array(self.getMean())
        np_variance = np.array(self.getStddev()) ** 2
        mask_value = self.getMaskValue()
        return StandardScaleLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            mean=np_mean,
            variance=np_variance,
            mask_value=mask_value,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class BaseLayer(tf.keras.layers.Layer, ABC):
    """
    Abstract base layer that performs casting of inputs and outputs to specified
    data types. All layers should inherit from this class.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the BaseLayer.

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: Input data type of the layer. If specified, inputs will be
        cast to this data type before any computation is performed. Defaults to `None`.
        :param output_dtype: Output data type of the layer. Defaults to `None`. If
        specified, the output will be cast to this data type before being returned.
        """
        super().__init__(name=name, **kwargs)
        # We handle casting of inputs and outputs in the call method
        # Allowing keras to also autocast causes issues in some layers that require
        # 64 bit precision. Such as timestamp layers after the year 2038.
        self._autocast = False
        # Needed to ensure keras 3 does not autocast inputs to float32
        self._convert_input_args = False
        self._input_dtype = input_dtype
        self._output_dtype = output_dtype
        self.true_bool_strings = ["true", "t", "yes", "y", "1"]
        self.false_bool_strings = ["false", "f", "no", "n", "0"]

    @property
    @abstractmethod
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        raise NotImplementedError

    def _string_to_bool_cast(self, inputs: Tensor) -> Tensor:
        """
        Casts a string tensor to a bool tensor.

        :param inputs: Input string tensor
        :returns: Bool tensor.
        """
        if inputs.dtype.name != "string":
            raise TypeError(
                f"Expected a string tensor, but got a {inputs.dtype.name} tensor."
            )

        # Replace true strings with "1" and false strings with "0"
        is_bool_true_string_tensor = [
            tf.strings.lower(inputs) == bool_string
            for bool_string in self.true_bool_strings
        ]
        is_bool_false_string_tensor = [
            tf.strings.lower(inputs) == bool_string
            for bool_string in self.false_bool_strings
        ]

        string_bool_tensor = tf.where(
            reduce(tf.math.logical_or, is_bool_true_string_tensor),
            tf.constant("1"),
            inputs,
        )
        string_bool_tensor = tf.where(
            reduce(tf.math.logical_or, is_bool_false_string_tensor),
            tf.constant("0"),
            string_bool_tensor,
        )

        # If we have other strings that are not "1" or "0", these are invalid.
        # We insert these as "NULL" values so that the casting will fail.
        string_bool_tensor_with_invalid = tf.where(
            tf.math.logical_or(string_bool_tensor == "1", string_bool_tensor == "0"),
            string_bool_tensor,
            tf.constant("NULL"),
        )

        bool_float_tensor = tf.strings.to_number(
            string_bool_tensor_with_invalid, out_type=tf.float32
        )
        return tf.cast(bool_float_tensor, tf.bool)

    @staticmethod
    def _float_to_string_cast(inputs: Tensor) -> Tensor:
        """
        Casts a float tensor to a string tensor. Ensures that the precision of the float
        does not impact the string representation. Specifically, we want the string
        to be the shortest possible representation of the float,
        i.e. 1.145000 -> "1.145".

        However, we also want to ensure that the string representation of the float
        has a decimal point, i.e. 2.00000 -> "2.0" and not "2".

        :param inputs: Input string tensor
        :returns: Float tensor.
        """
        # This gives 1.145000 -> "1.145" and 2.00000 -> "2".
        # We need to add a decimal point to the second example.
        shortest_float_string = tf.strings.as_string(inputs, shortest=True)

        # Find strings without decimal points
        no_decimal = tf.logical_not(
            tf.strings.regex_full_match(
                shortest_float_string, "-?\d*\.\d*"  # noqa W605
            )
        )
        # Create decimal point constant string
        decimal_string = tf.constant(".0")

        # Add decimal point to string without decimal points
        return tf.where(
            no_decimal,
            tf.strings.join([shortest_float_string, decimal_string]),
            shortest_float_string,
        )

    def _to_string_cast(self, inputs: Tensor) -> Tensor:
        """
        Casts inputs to string tensor.

        :param inputs: Input tensor.
        :returns: String tensor.
        """
        if inputs.dtype.is_floating:
            return self._float_to_string_cast(inputs)
        return tf.strings.as_string(inputs)

    def _from_string_cast(self, inputs: Tensor, cast_dtype: str) -> Tensor:
        """
        Casts inputs to the desired dtype when inputs are a string tensor.

        :param inputs: String tensor
        :param cast_dtype: Dtype to cast to.
        :returns: Tensor cast to the desired dtype.
        """
        if inputs.dtype.name != "string":
            raise TypeError("inputs is not a string Tensor.")
        if cast_dtype in ["float32", "float64", "int32", "int64"]:
            # If the casting dtype is supported by tf.strings.to_number, we use that.
            return tf.strings.to_number(inputs, out_type=cast_dtype)
        elif tf.as_dtype(cast_dtype).is_integer:
            # If the casting dtype is an integer, we need to cast to int64 first
            intermediate_cast = tf.strings.to_number(inputs, out_type="int64")
            return tf.cast(intermediate_cast, cast_dtype)
        elif tf.as_dtype(cast_dtype).is_floating:
            # If the casting dtype is a float, we need to cast to float64 first
            intermediate_cast = tf.strings.to_number(inputs, out_type="float64")
            return tf.cast(intermediate_cast, cast_dtype)
        elif tf.as_dtype(cast_dtype).is_bool:
            # If the casting dtype is a boolean, we need to use a custom function
            # to cast the string to boolean.
            return self._string_to_bool_cast(inputs)
        else:
            raise TypeError(f"Casting string to dtype {cast_dtype} is not supported.")

    def _string_cast(self, inputs: Tensor, cast_dtype: str) -> Tensor:
        """
        Casts from and to string tensors.

        Either inputs is a string tensor, and we want to cast it to the desired dtype,
        or inputs is not a string tensor, and we want to cast it to a string tensor.

        :param inputs: Input tensor.
        :param cast_dtype: Dtype to cast to.
        :returns: Tensor cast to the desired dtype.
        """
        if inputs.dtype.name == "string" and cast_dtype == "string":
            return inputs
        if cast_dtype == "string":
            return self._to_string_cast(inputs)
        return self._from_string_cast(inputs, cast_dtype)

    @staticmethod
    def _numeric_cast(inputs: Tensor, cast_dtype: str) -> Tensor:
        """
        Casts a numeric tensor to the desired (non-string) dtype.

        :param inputs: Input numeric tensor
        :param cast_dtype: Dtype to cast to.
        :returns: Tensor cast to the desired dtype.
        """
        return tf.cast(inputs, cast_dtype)

    def _cast(self, inputs: Tensor, cast_dtype: str) -> Tensor:
        """
        Casts inputs to the desired dtype.

        :param inputs: Input tensor.
        :param cast_dtype: Dtype to cast to.
        :returns: Tensor cast to the desired dtype.
        """
        if inputs.dtype.name == "string" or cast_dtype == "string":
            # If input tensor is a string tensor, or we are casting to a string,
            # we need to use the string_cast function.
            return self._string_cast(inputs, cast_dtype)
        else:
            return self._numeric_cast(inputs, cast_dtype)

    def _force_cast_to_compatible_numeric_type(
        self, inputs: Tensor, constant: Union[float, int]
    ) -> Tuple[Tensor, Tensor]:
        """
        Casts an input tensor and a single constant to compatible tensors.

        If the provided input is a float, create the constant tensor as a float of the
        same precision. If the provided input is an integer, check if the constant is
        non-floating, and if so, create the constant tensor as an integer of the same
        precision. If the constant is floating, cast the input to a float with the same
        precision as its integer dtype and create the constant tensor likewise.

        :param inputs: Input numeric tensor
        :param constant: The constant to cast to the compatible dtype.
        :returns: Tuple of tensors cast to compatible types
        """
        if inputs.dtype.is_floating:
            if isinstance(constant, float):
                return inputs, tf.constant(constant, dtype=inputs.dtype)
            return inputs, tf.constant(float(constant), dtype=inputs.dtype)
        if inputs.dtype.is_integer:
            if isinstance(constant, int):
                return inputs, tf.constant(constant, dtype=inputs.dtype)
            if isinstance(constant, float) and constant.is_integer():
                return inputs, tf.constant(int(constant), dtype=inputs.dtype)
            if isinstance(constant, float):
                precision = inputs.dtype.size * 8
                return (
                    self._cast(inputs, f"float{precision}"),
                    tf.constant(constant, dtype=f"float{precision}"),
                )
        raise TypeError(
            "inputs must be a numeric tensor and constant must be a numeric value."
        )

    def _cast_input_output_tensors(
        self, tensors: Union[Tensor, List[Tensor]], ingress: bool
    ) -> Union[Tensor, List[Tensor]]:
        """
        Casts either the input or output tensors to the given input/output dtype, if
        specified. Ingress is a boolean that indicates whether we are casting the
        input (True) or output (False) tensors.

        :param tensors: The input or output tensor(s) to the layer to be cast.
        :param ingress: Boolean indicating whether we are casting the input (True) or
        output (False) tensors.
        :returns: The input or output tensor(s) cast to the desired input/output_dtype.
        """
        if ingress:
            cast_dtype = self._input_dtype
            if (
                cast_dtype is not None
                and self.compatible_dtypes is not None
                and cast_dtype not in [dtype.name for dtype in self.compatible_dtypes]
            ):
                raise ValueError(
                    f"""input_dtype {cast_dtype} is not a compatible dtype for
                    this layer. Compatible dtypes are {[
                        dtype.name for dtype in self.compatible_dtypes
                    ]}."""
                )
        else:
            cast_dtype = self._output_dtype

        if cast_dtype is not None:
            if tf.is_tensor(tensors):
                return (
                    self._cast(tensors, cast_dtype)
                    if tensors.dtype.name != cast_dtype
                    else tensors
                )
            return [
                self._cast(inp, cast_dtype) if inp.dtype.name != cast_dtype else inp
                for inp in tensors
            ]
        return tensors

    def cast_input_tensors(
        self, inputs: Union[Tensor, List[Tensor]]
    ) -> Union[Tensor, List[Tensor]]:
        """
        Casts the input tensors to the given input dtype, if specified. All tensors are
        cast to this. This might not be ideal, there may be layers where some inputs are
        expected to be different types. In these cases, the subclass should
        implement the cast_input_tensors method.

        :param inputs: The input tensor(s) to the layer.
        :returns: The input tensor(s) cast to the desired input_dtype.
        """
        return self._cast_input_output_tensors(tensors=inputs, ingress=True)

    def cast_output_tensors(
        self, outputs: Union[Tensor, List[Tensor]]
    ) -> Union[Tensor, List[Tensor]]:
        """
        Casts the output tensors to the given output dtype, if specified. All tensors
        are cast to this. This might not be ideal, there may be layers where some
        outputs are expected to be different types. In these cases, the subclass should
        implement the cast_output_tensors method.

        :param outputs: The output tensor(s) of the layer.
        :returns: The output tensor(s) cast to the desired output_dtype.
        """
        return self._cast_input_output_tensors(tensors=outputs, ingress=False)

    def _check_input_dtypes_compatible(self, inputs: List[Tensor]) -> None:
        """
        Checks if the input tensors are compatible with the compatible_dtypes of the
        layer.

        :param inputs: The input tensor(s) to the layer.
        :raises ValueError: If the input tensors are not compatible with the
        compatible_dtypes of the layer.
        :returns: None
        """
        for inp in inputs:
            if (
                self.compatible_dtypes is not None
                and inp.dtype not in self.compatible_dtypes
            ):
                raise TypeError(
                    f"""Input tensor with dtype {inp.dtype.name}
                    is not a compatible dtype for this layer.
                    Compatible dtypes are {[
                        dtype.name for dtype in self.compatible_dtypes
                    ]}."""
                )

    @allow_single_or_multiple_tensor_input
    def call(self, inputs: Iterable[Tensor], **kwargs) -> Union[Tensor, List[Tensor]]:
        """
        Casts inputs to the given `input_dtype`, calls the internal `_call` method, and
        casts the outputs to the given `output_dtype`.

        :param inputs: The input tensor(s) to the layer.
        :returns: The output tensor(s) of the layer.
        """
        # Cast inputs to a compatible dtype for the layer
        casted_inputs = self.cast_input_tensors(inputs=inputs)
        # Check if the input tensors are now compatible with the layer
        self._check_input_dtypes_compatible(inputs=casted_inputs)
        # Call the internal _call method
        outputs = self._call(inputs=casted_inputs, **kwargs)
        # Cast outputs to the desired output_dtype
        casted_outputs = self.cast_output_tensors(outputs=outputs)
        return casted_outputs

    @abstractmethod
    def _call(
        self, inputs: Union[Tensor, List[Tensor]], **kwargs
    ) -> Union[Tensor, List[Tensor]]:
        """
        The internal call method that should be implemented by the layer.

        :param inputs: The input tensor(s) to the layer.
        :returns: The output tensor(s) of the layer.
        """
        raise NotImplementedError

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the BaseLayer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "name": self.name,
                "input_dtype": self._input_dtype,
                "output_dtype": self._output_dtype,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import pandas as pd
import tensorflow as tf

from kamae.sklearn.params import SingleInputMultiOutputMixin
from kamae.tensorflow.layers import ArraySplitLayer

from .base import BaseTransformer


class ArraySplitTransformer(
    BaseTransformer,
    SingleInputMultiOutputMixin,
):
    """
    VectorSlicer Scikit-Learn Transformer for use in Scikit-Learn pipelines.
    This transformer slices an array column into multiple columns.
    """

    def __init__(self, input_col: str, output_cols: List[str], layer_name: str) -> None:
        super().__init__()
        self.input_col = input_col
        self.output_cols = output_cols
        self.layer_name = layer_name

    def fit(self, X: pd.DataFrame, y=None) -> "ArraySplitTransformer":
        """
        Fits the transformer to the data. Does nothing since
        this is transformer not an estimator.

        :param X: Pandas dataframe to fit the transformer to.
        :param y: Not used, present here for API consistency by convention.
        :returns: Fit pipeline, in this case the transformer itself.
        """
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transforms the input dataset. Creates a new column for each output column equal
        to the value of the input column at the given index.

        :param X: Pandas dataframe to transform.
        :param y: Not used, present here for API consistency by convention.
        :returns: Transformed data.
        """
        X[self.output_cols] = pd.DataFrame(X[self.input_col].tolist(), index=X.index)
        return X

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for that unstacks the input tensor and reshapes
        to the original shape.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that slices the input tensors.
        """
        return ArraySplitLayer(name=self.layer_name, axis=-1)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql.types import DataType

from kamae.spark.utils import (
    build_udf_return_type,
    flatten_nested_arrays,
    get_array_nesting_level_and_element_dtype,
    single_input_single_output_array_transform,
)


def construct_nested_elements_for_scaling(
    column: Column,
    column_datatype: DataType,
    array_dim: int,
) -> Column:
    """
    Scaling nested elements in Spark is difficult and requires us to extract the `ith`
    element from the innermost array to compute the moments on.
    This function creates multiple columns, one for each dimension of the inner array
    and flattens them into a single array. It then explodes these out so that the mean
    and variance can be computed on the flattened array.

    Only intended to be used for the StandardScaleEstimator &
    ConditionalStandardScaleEstimator.

    :param column: The input column to extract the element from.
    :param column_datatype: The datatype of the input column.
    :param array_dim: The dimension of the innermost array.
    :returns: A column containing a struct of (possibly exploded) elements.
    """
    nested_lvl, element_dtype = get_array_nesting_level_and_element_dtype(
        column_datatype
    )
    elements = [
        single_input_single_output_array_transform(
            input_col=column,
            input_col_datatype=column_datatype,
            func=lambda x: F.element_at(x, idx),
        ).alias(f"element_{idx}")
        for idx in range(1, array_dim + 1)
    ]
    if nested_lvl > 1:
        # If the element is a nested array, we need to flatten it.
        flat_elements = [
            flatten_nested_arrays(
                column=e,
                column_data_type=build_udf_return_type(
                    element_dtype=element_dtype, nest_level=nested_lvl - 1
                ),
            )
            for e in elements
        ]
        # We then zip the flattened elements together and call explode.
        # Explode is needed in order to use the spark mean and stddev functions
        # over the arrays.
        return F.explode(F.arrays_zip(*flat_elements)).alias("element_struct")

    else:
        # If the element is scalar then we wrap it in a struct to mimic
        # the same structure as the nested arrays.
        return F.struct(*elements).alias("element_struct")



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import (
    ArrayType,
    DataType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
)

from kamae.spark.params import (
    DropUnseenParams,
    MultiInputMultiOutputParams,
    StringIndexParams,
)
from kamae.spark.utils import (
    one_hot_encoding_udf,
    single_input_single_output_scalar_udf_transform,
)
from kamae.tensorflow.layers import OneHotEncodeLayer

from .base import BaseTransformer


class SharedOneHotEncodeTransformer(
    BaseTransformer,
    MultiInputMultiOutputParams,
    StringIndexParams,
    DropUnseenParams,
):
    """
    SharedOneHotEncodeTransformer Spark Transformer for use in Spark pipelines.
    This transformer is used to one-hot encode multiple input columns using the
    string labels collected by the SharedOneHotEncodeEstimator.

    NOTE: If your data contains null characters:
    https://en.wikipedia.org/wiki/Null_character
    This transformer could fail since the hashing algorithm uses cannot accept null
    characters. If you have null characters in your data, you should remove them.
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCols: Optional[List[str]] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        labelsArray: Optional[List[str]] = None,
        stringOrderType: str = "frequencyDesc",
        maskToken: Optional[str] = None,
        numOOVIndices: int = 1,
        dropUnseen: bool = False,
    ) -> None:
        """
        Initializes the SharedOneHotEncodeTransformer transformer.

        :param inputCols: List of input column names.
        :param outputCols: List of output column name.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param labelsArray: List of string labels to use for one-hot encoding.
        :param: stringOrderType: How to order the string indices.
        Options are 'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc',
        'alphabeticalDesc'. Defaults to 'frequencyDesc'.
        :param maskToken: Token to use for masking.
        If set, the token will be indexed as 0.
        :param numOOVIndices: Number of out of vocabulary indices to use. The
        out of vocabulary indices are used to represent unseen labels and are
        placed at the beginning of the one-hot encoding. Defaults to 1.
        :param dropUnseen: Whether to drop unseen label indices. If set to True,
        the transformer will not add an extra dimension for unseen labels in the
        one-hot encoding. Defaults to False.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            stringOrderType="frequencyDesc",
            numOOVIndices=1,
            dropUnseen=False,
            maskToken=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [ShortType(), IntegerType(), LongType(), StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset using the string index labels assigning array of
        one-hot encoded values to the output columns.

        :param dataset: Pyspark dataframe to transform.

        :returns: Pyspark dataframe with the input column one-hot encoded,
         named as the output column.
        """
        labels = self.getLabelsArray()
        ohe_num_oov_indices = self.getNumOOVIndices()
        mask_token = self.getMaskToken()
        drop_unseen = self.getDropUnseen()

        # Assumption made that all the input columns have the same datatype/nesting.
        input_datatypes = [
            self.get_column_datatype(dataset=dataset, column_name=c)
            for c in self.getInputCols()
        ]

        output_cols = []
        for i, column in enumerate(self.getInputCols()):
            output_col = single_input_single_output_scalar_udf_transform(
                input_col=F.col(column),
                input_col_datatype=input_datatypes[i],
                func=lambda x: one_hot_encoding_udf(
                    label=x,
                    labels=labels,
                    num_oov_indices=ohe_num_oov_indices,
                    mask_token=mask_token,
                    drop_unseen=drop_unseen,
                ),
                udf_return_element_datatype=ArrayType(FloatType()),
            )
            output_cols.append(output_col.alias(self.getOutputCols()[i]))

        select_cols = [F.col(c) for c in dataset.columns] + output_cols

        return dataset.select(*select_cols)

    def get_tf_layer(self) -> List[tf.keras.layers.Layer]:
        """
        Gets the list of tensorflow layers for the shared onehot encoder transformer.
        We need to use a list as each layer could operate on differing input shapes.

        :returns: List of Tensorflow keras layer with name equal to the layerName
        parameter and the input column name, that performs the indexing.
        """
        return [
            OneHotEncodeLayer(
                name=f"{self.getLayerName()}_{input_name}",
                input_dtype=self.getInputTFDtype(),
                output_dtype=self.getOutputTFDtype(),
                vocabulary=self.getLabelsArray(),
                num_oov_indices=self.getNumOOVIndices(),
                mask_token=self.getMaskToken(),
                drop_unseen=self.getDropUnseen(),
            )
            for input_name in self.getInputCols()
        ]



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
import base64
from typing import Any, Callable, List, Optional, Union

import dill
import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, StructField, StructType

from kamae.spark.params import (
    MultiInputMultiOutputParams,
    MultiInputSingleOutputParams,
    SingleInputMultiOutputParams,
    SingleInputSingleOutputParams,
)
from kamae.tensorflow.layers import LambdaFunctionLayer
from kamae.tensorflow.typing import Tensor

from .base import BaseTransformer


class LambdaFunctionParams(Params):
    """
    Mixin class containing parameters needed for the LambdaFunctionTransformer.
    """

    function = Param(
        Params._dummy(),
        "function",
        "Function to apply to the input column. Serialized by dill.",
        typeConverter=TypeConverters.toString,
    )

    functionReturnTypes = Param(
        Params._dummy(),
        "functionReturnTypes",
        """List of return types of the lambda function. Should be length 1 for
        single output. Used to understand the UDF return type.""",
        typeConverter=TypeConverters.toListString,
    )

    def setFunction(self, value: Callable[[Any], Any]) -> "LambdaFunctionParams":
        """
        Sets the lambda function to apply to the input column.

        :param value: Lambda function to apply to the input column.
        :returns: Class instance.
        """
        dill_bytes = dill.dumps(value)
        return self._set(function=base64.b64encode(dill_bytes).decode("utf-8"))

    def setFunctionReturnTypes(self, value: List[DataType]) -> "LambdaFunctionParams":
        """
        Sets the return type of the lambda function.

        :param value: Return type of the lambda function.
        :returns: Class instance.
        """
        serialised_dtypes = []
        for return_type in value:
            dill_bytes = dill.dumps(return_type)
            serialised_dtypes.append(base64.b64encode(dill_bytes).decode("utf-8"))

        return self._set(functionReturnTypes=serialised_dtypes)

    def getFunction(self) -> Callable[[Any], Any]:
        """
        Gets the lambda function to apply to the input column.

        :returns: Lambda function to apply to the input column.
        """
        utf_str = self.getOrDefault(self.function)
        return dill.loads(base64.b64decode(utf_str))

    def getFunctionReturnTypes(self) -> List[DataType]:
        """
        Gets the return type of the lambda function.

        :returns: Return type of the lambda function.
        """
        utf_str_list = self.getOrDefault(self.functionReturnTypes)
        return [dill.loads(base64.b64decode(utf_str)) for utf_str in utf_str_list]


class LambdaFunctionTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    SingleInputMultiOutputParams,
    MultiInputSingleOutputParams,
    MultiInputMultiOutputParams,
    LambdaFunctionParams,
):
    """
    Spark Transformer that applies tensorflow lambda functions to the input column(s).

    The provided function must either take a single tensor as input or take a list of
    tensors as input. Same for the output, either single tensor or list of tensors.
    In the case of multiple input/output, it is assumed that the order matches the
    order of the input/output columns.

    It can only use tensorflow methods, and reference them with `tf.`.

    An example of a valid lambda function is:

    ```python
    def my_tf_fn(x):
        return tf.square(x) - tf.math.log(x)
    ```

    Note: No validation is done on the lambda function. It is up to the user to
    ensure the lambda function is valid and works as expected. The lambda function
    must be serializable by dill: https://github.com/uqfoundation/dill

    The function will be used within a User Defined Function (UDF) in Spark.
    You need to provide the return type for the UDF computation as a string.
    E.g. "float", "array<string>" etc. If you require the operation to be quicker
    (in Spark), consider composing it from other transformers, that are written in
    native Spark functions.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        outputCols: Optional[List[str]] = None,
        function: Optional[
            Callable[[Union[Tensor, List[Tensor]]], Union[Tensor, List[Tensor]]]
        ] = None,
        functionReturnTypes: Optional[List[DataType]] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes an LambdaFunctionTransformer transformer.

        :param inputCol: Input column name.
        :param inputCols: List of input column names. Used for multiple input columns.
        inputCols and inputCol cannot be set at the same time.
        :param outputCol: Output column name.
        :param outputCols: List of output column names. Used for multiple output
        columns. outputCols and outputCol cannot be set at the same time.
        :param function: Lambda function to apply to the input column. If single input,
        the function should take a single tensor as input and return a single tensor.
        If multiple input, the function should take a list of tensors as input and
        return a single tensor.
        :param functionReturnTypes: Return type(s) of the lambda function. List of
        Pyspark datatypes. Used to understand the UDF return type in Spark.
        Keras layer does not use this. E.g. "float", "array<string>" etc. If outputCol
        is set, this should be a list of length 1. If outputCols is set, this should be
        a list of the same length as outputCols.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def setInputCol(self, value: str) -> "LambdaFunctionTransformer":
        """
        Sets the parameter inputCol to the given string value.
        Only allows setting if inputCols is not set.

        :param value: String to set the inputCol parameter to.
        :returns: Instance of class mixed in.
        """
        if self.isSet("inputCols"):
            raise ValueError("inputCols is already set. Cannot set inputCol as well.")
        return self._set(inputCol=value)

    def setInputCols(self, value: List[str]) -> "LambdaFunctionTransformer":
        """
        Sets the parameter inputCols to the given list of strings.
        Only allows setting if inputCol is not set. Only allows multiple inputs.

        :param value: List of strings to set the inputCols parameter to.
        :returns: Instance of class mixed in.
        """
        if len(value) < 2:
            raise ValueError(
                """inputCols must be set with more than one column.
                Use inputCol for single input."""
            )
        if self.isSet("inputCol"):
            raise ValueError("inputCol is already set. Cannot set inputCols as well.")
        return self._set(inputCols=value)

    def setOutputCol(self, value: str) -> "LambdaFunctionTransformer":
        """
        Sets the parameter outputCol to the given string value.
        Only allows setting if outputCols is not set.

        :param value: String to set the outputCol parameter to.
        :returns: Instance of class mixed in.
        """
        if self.isSet("outputCols"):
            raise ValueError("outputCols is already set. Cannot set outputCol as well.")
        return self._set(outputCol=value)

    def setOutputCols(self, value: List[str]) -> "LambdaFunctionTransformer":
        """
        Sets the parameter outputCols to the given list of strings.
        Only allows setting if outputCol is not set. Only allows multiple outputs.

        :param value: List of strings to set the outputCols parameter to.
        :returns: Instance of class mixed in.
        """
        if len(value) < 2:
            raise ValueError(
                """outputCols must be set with more than one column.
                Use outputCol for single output."""
            )
        if self.isSet("outputCol"):
            raise ValueError("outputCol is already set. Cannot set outputCols as well.")
        return self._set(outputCols=value)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _validate_params(self) -> None:
        """
        Validates the parameters of the LambdaFunctionTransformer.
        """
        if not (self.isDefined("function") and self.isDefined("functionReturnTypes")):
            raise ValueError("function and functionReturnTypes must be set.")

        if not (self.isDefined("inputCol") or self.isDefined("inputCols")):
            raise ValueError("inputCol or inputCols must be set.")

        if not (self.isDefined("outputCol") or self.isDefined("outputCols")):
            raise ValueError("outputCol or outputCols must be set.")

        output_col_names = self._get_output_cols()
        if len(output_col_names) != len(self.getFunctionReturnTypes()):
            raise ValueError(
                "Number of output columns must match number of function return types."
            )

    def _get_input_cols(self) -> List[str]:
        """
        Gets the input columns for the transformer.

        :returns: List of input column names.
        """
        return (
            self.getInputCols() if self.isDefined("inputCols") else [self.getInputCol()]
        )

    def _get_output_cols(self) -> List[str]:
        """
        Gets the output columns for the transformer.

        :returns: List of output column names.
        """
        return (
            self.getOutputCols()
            if self.isDefined("outputCols")
            else [self.getOutputCol()]
        )

    @staticmethod
    def _apply_udf_func_to_dataset(
        dataset: DataFrame,
        func: Callable[[Union[Tensor, List[Tensor]]], Any],
        input_col_names: List[str],
        output_col_names: List[str],
        function_return_types: List[DataType],
    ) -> DataFrame:
        """
        Gets and applies the udf to the dataset. If the output is a single column, the
        output is directly applied to the dataset. If the output is multiple columns,
        a struct column is created and then the columns are extracted.

        :param dataset: Pyspark dataframe to transform.
        :param func: Tensorflow function.
        :param input_col_names: List of input column names.
        :param output_col_names: List of output column names.
        :param function_return_types: List of return types of the lambda function.
        :returns: Transformed pyspark dataframe.
        """
        if len(output_col_names) == 1:
            # Single output scenario
            output_col = F.udf(func, function_return_types[0])(
                *[F.col(c) for c in input_col_names]
            )
            return dataset.withColumn(output_col_names[0], output_col)
        else:
            # Multi-output scenario. UDF cannot return multiple columns, so
            # return a struct type and then extract the columns.
            udf_return_type = StructType(
                [
                    StructField(name=output_col, dataType=function_return_types[i])
                    for i, output_col in enumerate(output_col_names)
                ]
            )
            output_struct_col = F.udf(func, udf_return_type)(
                *[F.col(c) for c in input_col_names]
            )
            for output_col in output_col_names:
                dataset = dataset.withColumn(output_col, output_struct_col[output_col])
            return dataset

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which applies `function` to the input column(s).

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        self._validate_params()

        input_col_names = self._get_input_cols()
        output_col_names = self._get_output_cols()
        function_return_types = self.getFunctionReturnTypes()

        def tf_function_wrapper(
            fn: Callable[[Union[Tensor, List[Tensor]]], Tensor],
        ) -> Callable[[Union[Tensor, List[Tensor]]], Any]:
            """
            Wraps the tensorflow function, so we can use it within a Spark UDF.

            Specifically:

            - Handles the fact that Spark UDFs require multiple inputs to be passed as
            arguments, not in a list. However, the tensorflow function expects multiple
            inputs to be passed as a list.
            - Wraps scalar inputs in a list before creating the tensor. The tensor also
            has a single batch dimension. Therefore, a scalar value has shape (1, 1)
            and an array has shape (1, N).
            - Converts the output tensor to a numpy value and extracts the single batch.
            If value is a list of size 1, return the single value.
            - If the output tensor is a string, decodes the bytes to a string.

            :param fn: Tensorflow function.
            :returns: Function that can be used within a Spark UDF.
            """

            def wrapper(*args):
                # Wrap args in a list if they are not already a list. This ensures
                # that scalar inputs have shape (1, 1) in tensorflow.
                args = [a if isinstance(a, list) else [a] for a in args]
                # Call the function with the input tensors.
                if len(args) == 1:
                    tf_val: Union[Tensor, list[Tensor]] = fn(tf.constant([args[0]]))
                else:
                    tf_val: Union[Tensor, list[Tensor]] = fn(
                        [tf.constant([a]) for a in args]
                    )

                # Single output scenario
                if len(function_return_types) == 1:
                    if "string" in function_return_types[0].simpleString():
                        # If string output, decode the bytes to a string.
                        output = tf_val.numpy().astype(str).tolist()[0]
                    else:
                        output = tf_val.numpy().tolist()[0]
                    if len(output) == 1 and not isinstance(
                        function_return_types[0], ArrayType
                    ):
                        return output[0]
                    return output
                # Multi output scenario
                else:
                    tf_outputs = []
                    for i, tensor in enumerate(tf_val):
                        if "string" in function_return_types[i].simpleString():
                            # If string output, decode the bytes to a string.
                            output = tensor.numpy().astype(str).tolist()[0]
                        else:
                            output = tensor.numpy().tolist()[0]
                        if len(output) == 1 and not isinstance(
                            function_return_types[i], ArrayType
                        ):
                            # If the output is a scalar value, extract it from the list.
                            tf_outputs.append(output[0])
                        else:
                            tf_outputs.append(output)
                    # Return a tuple, which will be unpacked into multiple columns.
                    return tuple(tf_outputs)

            return wrapper

        func = tf_function_wrapper(self.getFunction())

        return self._apply_udf_func_to_dataset(
            dataset=dataset,
            func=func,
            input_col_names=input_col_names,
            output_col_names=output_col_names,
            function_return_types=function_return_types,
        )

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the lambda function transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs the lambda function on the input.
        """
        return LambdaFunctionLayer(
            function=self.getFunction(),
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional, Union

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType

from kamae.spark.params import (
    MultiInputSingleOutputParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import IfStatementLayer
from kamae.utils import get_condition_operator

from .base import BaseTransformer


class IfStatementParams(Params):
    """
    Mixin class containing parameters needed for IfStatementTransformer
    transform layers.
    """

    conditionOperator = Param(
        Params._dummy(),
        "conditionOperator",
        "Operator to use in condition: eq, neq, lt, gt, leq, geq",
        typeConverter=TypeConverters.toString,
    )

    valueToCompare = Param(
        Params._dummy(),
        "valueToCompare",
        "Value to compare to input column",
        typeConverter=TypeConverters.identity,
    )

    resultIfTrue = Param(
        Params._dummy(),
        "resultIfTrue",
        "Value to return if condition is true",
        typeConverter=TypeConverters.identity,
    )

    resultIfFalse = Param(
        Params._dummy(),
        "resultIfFalse",
        "Value to return if condition is false",
        typeConverter=TypeConverters.identity,
    )

    def setConditionOperator(self, value: str) -> "IfStatementParams":
        """
        Sets the conditionOperator parameter.

        :param value: String value describing the operator to use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        :returns: Instance of class mixed in.
        """
        allowed_operators = ["eq", "neq", "lt", "gt", "leq", "geq"]
        if value not in allowed_operators:
            raise ValueError(
                f"conditionOperator must be one of {allowed_operators}, but got {value}"
            )
        if (
            self.getValueToCompare() is not None
            and not isinstance(self.getValueToCompare(), Number)
            and value in ["lt", "leq", "gt", "geq"]
        ):
            raise ValueError(
                """conditionOperator must be eq or neq when
                valueToCompare is not a number."""
            )
        return self._set(conditionOperator=value)

    def setResultIfTrue(
        self, value: Union[float, int, str, bool]
    ) -> "IfStatementParams":
        """
        Sets the resultIfTrue parameter.

        :param value: Value to return if condition is true.
        :returns: Instance of class mixed in.
        """
        return self._set(resultIfTrue=value)

    def setResultIfFalse(
        self, value: Union[float, int, str, bool]
    ) -> "IfStatementParams":
        """
        Sets the resultIfFalse parameter.

        :param value: Value to return if condition is false.
        :returns: Instance of class mixed in.
        """
        return self._set(resultIfFalse=value)

    def setValueToCompare(
        self, value: Union[float, int, str, bool]
    ) -> "IfStatementParams":
        """
        Sets the valueToCompare parameter.

        :param value: Value to compare to input column.
        :returns: Instance of class mixed in.
        """
        if (
            self.isDefined("conditionOperator")
            and not isinstance(value, Number)
            and self.getConditionOperator() in ["lt", "leq", "gt", "geq"]
        ):
            raise TypeError(
                """valueToCompare must be a number if conditionOperator is one of
                lt, leq, gt, or geq."""
            )
        return self._set(valueToCompare=value)

    def getConditionOperator(self) -> str:
        """
        Gets the conditionOperator parameter.

        :returns: String value describing the operator to use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        """
        return self.getOrDefault(self.conditionOperator)

    def getValueToCompare(self) -> Union[float, int, str, bool]:
        """
        Gets the valueToCompare parameter.

        :returns: Value to compare to input column.
        """
        return self.getOrDefault(self.valueToCompare)

    def getResultIfTrue(self) -> Union[float, int, str, bool]:
        """
        Gets the resultIfTrue parameter.

        :returns: Value to return if condition is true.
        """
        return self.getOrDefault(self.resultIfTrue)

    def getResultIfFalse(self) -> Union[float, int, str, bool]:
        """
        Gets the resultIfFalse parameter.

        :returns: Value to return if condition is false.
        """
        return self.getOrDefault(self.resultIfFalse)


class IfStatementTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    IfStatementParams,
):
    """
    IfStatement Spark Transformer for use in Spark pipelines.
    This transformer computes an if statement between a set of constants
    and columns.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        conditionOperator: Optional[str] = None,
        valueToCompare: Optional[Union[float, int, str, bool]] = None,
        resultIfTrue: Optional[Union[float, int, str, bool]] = None,
        resultIfFalse: Optional[Union[float, int, str, bool]] = None,
    ) -> None:
        """
        Initializes an IfStatementTransformer transformer.

        :param inputCol: Input column name. Only used if inputCols is not specified.
        If specified, then all other aspects of the if statement are constant.
        :param inputCols: Input column names. List of input columns to in the case
        where the if statement is not constant. Must be specified in the order
        [valueToCompare, resultIfTrue, resultIfFalse].
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param conditionOperator: Operator to use in condition:
        eq, neq, lt, gt, leq, geq.
        :param valueToCompare: Optional value to compare to input column.
        If not specified, then assumed to be the first input column.
        :param resultIfTrue: Optional value to return if condition is true.
        If not specified, then assumed to be the second input column.
        :param resultIfFalse: Optional value to return if condition is false.
        If not specified, then assumed to be the third input column.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            valueToCompare=None,
            resultIfTrue=None,
            resultIfFalse=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def setInputCols(self, value: List[str]) -> "IfStatementTransformer":
        """
        Sets the inputCols parameter, throwing an error if the total length of the
        inputCols and the constants are more than 4.

        :param value: List of input column names.
        :returns: Instance of class mixed in.
        """
        optional_constant_cols = ["valueToCompare", "resultIfTrue", "resultIfFalse"]
        num_defined_constants = len(
            [c for c in optional_constant_cols if self.getOrDefault(c) is not None]
        )
        if len(value) + num_defined_constants > 4:
            raise ValueError(
                f"""Total number of input columns and constants cannot be more than 4,
                but got {len(value)} input columns and
                {num_defined_constants} constants."""
            )

        return self._set(inputCols=value)

    def _construct_input_cols(self) -> List[Column]:
        """
        Constructs the input columns for the transformer. This is fairly complex
        since the user can set any of the following:
        - valueToCompare
        - resultIfTrue
        - resultIfFalse

        If all of these are set, then the user should have provided a single `inputCol`.
        Otherwise, if any of these are not set, then the user should have provided
        `inputCols` containing the missing values. The `inputCols` should be in the
        order [valueToCompare, resultIfTrue, resultIfFalse]. But if the user has
        specified, for example, `resultIfTrue` as a constant then the
        `inputCols` should be in the order [valueToCompare, resultIfFalse].

        :returns: Tuple of 4 pyspark columns.
        """
        optional_constant_cols = ["valueToCompare", "resultIfTrue", "resultIfFalse"]
        optional_constants_defined = {
            const: self.getOrDefault(const) is not None
            for const in optional_constant_cols
        }

        if self.isDefined("inputCols"):
            # If the user has set inputCols, then some or all of the optional constants
            # are defined as input column variables
            input_cols = self.getInputCols()
            if len(input_cols) + sum(optional_constants_defined.values()) != 4:
                raise ValueError(
                    f"""Total number of input columns and constants must be 4,
                    but got {len(input_cols)} input columns and
                    {sum(optional_constants_defined.values())} constants."""
                )
            # The first input column is always the value to compare
            input_col_list = [
                F.col(input_cols[0]),
            ]
            input_col_counter = 1
            for const_col in optional_constant_cols:
                # Loop through the optional constant names.
                # If the constant is not defined then it must be an input column.
                # Otherwise, it is a literal value.
                if not optional_constants_defined[const_col]:
                    input_col_list.append(F.col(input_cols[input_col_counter]))
                    input_col_counter += 1
                else:
                    input_col_list.append(F.lit(self.getOrDefault(const_col)))
            return [
                input_col_list[0].alias(self.uid + "_inputCol"),
                input_col_list[1].alias(self.uid + "_valueToCompare"),
                input_col_list[2].alias(self.uid + "_resultIfTrue"),
                input_col_list[3].alias(self.uid + "_resultIfFalse"),
            ]
        elif self.isDefined("inputCol"):
            # If the user has set inputCol, then all the optional constants
            # must be defined.
            if not all(
                [
                    self.getOrDefault(const) is not None
                    for const in optional_constant_cols
                ]
            ):
                raise ValueError(
                    f"""Must specify all of {optional_constant_cols}"
                    if inputCol is specified."""
                )
            return [
                F.col(self.getInputCol()).alias(self.uid + "_inputCol"),
                F.lit(self.getValueToCompare()).alias(self.uid + "_valueToCompare"),
                F.lit(self.getResultIfTrue()).alias(self.uid + "_resultIfTrue"),
                F.lit(self.getResultIfFalse()).alias(self.uid + "_resultIfFalse"),
            ]
        else:
            raise ValueError("Must specify either inputCol or inputCols.")

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the output of an if statement with condition:

        IF `inputCol` `conditionOperator` `valueToCompare`
        THEN `resultIfTrue`
        ELSE `resultIfFalse`

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self._construct_input_cols()
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]
        condition_operator = get_condition_operator(self.getConditionOperator())

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: F.when(
                condition_operator(x[input_col_names[0]], x[input_col_names[1]]),
                x[input_col_names[2]],
            ).otherwise(x[input_col_names[3]]),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the numerical if statement transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs the numerical if statement.
        """
        if not self.isDefined("conditionOperator"):
            raise ValueError("Must specify conditionOperator to use tensorflow layer.")

        return IfStatementLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            condition_operator=self.getConditionOperator(),
            value_to_compare=self.getValueToCompare(),
            result_if_true=self.getResultIfTrue(),
            result_if_false=self.getResultIfFalse(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import datetime_total_days, enforce_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class DateDiffLayer(BaseLayer):
    """A preprocessing layer that returns the difference between two dates in days.

    The inputs must be in yyyy-MM-dd (HH:mm:ss.SSS) format and
    must be passed to the layer in the order [start date , end date].
    The transformer will return a negative value if the order is reversed.
    """

    def __init__(
        self,
        name=None,
        input_dtype: str = None,
        output_dtype: str = None,
        default_value: Optional[int] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the DateDiffLayer layer.

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.default_value = default_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_multiple_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the date difference operation on two input tensors.

        Decorated with `@enforce_multiple_tensor_input` to ensure that the input
        is an iterable. Raises an error if a single tensor is passed.

        We also then check if the length of the iterable is 2.
        If not, we raise an error.

        :param inputs: Iterable of two tensors to perform the date difference operation
        on.
        :returns: Single tensor with the difference between the two dates in days.
        """
        if len(inputs) != 2:
            raise ValueError("Input shape must be an iterable of two tensors")

        start_date, end_date = inputs
        if self.default_value is not None:
            # Trick to replace empty strings with a valid dummy date, that we ignore
            # later. Otherwise, the date_difference function will raise an error
            replaced_start_date = tf.where(
                tf.equal(start_date, ""), "2000-01-01 00:00:00.000", start_date
            )
            replaced_end_date = tf.where(
                tf.equal(end_date, ""), "2000-01-01 00:00:00.000", end_date
            )
            outputs = tf.where(
                tf.logical_or(tf.equal(start_date, ""), tf.equal(end_date, "")),
                tf.constant(self.default_value, dtype=tf.int64),
                self.date_difference(replaced_end_date, replaced_start_date),
            )
        else:
            outputs = self.date_difference(end_date, start_date)
        return outputs

    def date_difference(self, end_date: Tensor, start_date: Tensor) -> Tensor:
        """
        Calculates the difference between two dates.

        :param end_date: Tensor of end dates.
        :param start_date: Tensor of start dates.
        :returns: Tensor of date difference in days.
        """
        return datetime_total_days(end_date) - datetime_total_days(start_date)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the DateDiff layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"default_value": self.default_value})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ExpLayer(BaseLayer):
    """
    Performs the exp(x) operation on a given input tensor
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initializes the exp layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.complex64,
            tf.complex128,
        ]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the exp(x) operation on a given input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Tensor to perform the exp(x) operation on.
        :returns: The exp of the input tensor.
        """
        return tf.math.exp(inputs)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the exp layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import (
    ConstantStringArrayParams,
    NegationParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringIsInListLayer

from .base import BaseTransformer


class StringIsInListTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    NegationParams,
    ConstantStringArrayParams,
):
    """
    String is in list Spark Transformer for use in Spark pipelines.
    This transformer performs a string equality operation on the input column over all
    constants in the passed constantStringArray.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        constantStringArray: Optional[List[str]] = None,
        negation: bool = False,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Intializes a StringIsInListTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param constantStringArray: String constant array to use in string isin list
        operation.
        :param negation: Whether to negate the string isin list operation.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(negation=False)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which contains the result of the string isin operation.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        if not self.isDefined("constantStringArray"):
            raise ValueError("constantStringArray must be defined.")

        def string_isin_list(
            x: Column, string_list: List[str], negation: bool
        ) -> Column:
            col_expr = x.isin(string_list)
            return col_expr if not negation else ~col_expr

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: string_isin_list(
                x=x,
                string_list=self.getConstantStringArray(),
                negation=self.getNegation(),
            ),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the StringIsInListLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
        performs a string isin operation.
        """

        if not self.isDefined("constantStringArray"):
            raise ValueError("constantStringArray must be defined.")

        return StringIsInListLayer(
            name=self.getLayerName(),
            negation=self.getNegation(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            string_constant_list=self.getConstantStringArray(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import Any, List, Optional, Union

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import (
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
)

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import BinLayer
from kamae.utils import get_condition_operator

from .base import BaseTransformer


class BinParams(Params):
    """
    Mixin class containing parameters needed for Bin transform layers.
    """

    conditionOperators = Param(
        Params._dummy(),
        "conditionOperators",
        "Operators to use in condition: eq, neq, lt, gt, leq, geq",
        typeConverter=TypeConverters.toListString,
    )

    binValues = Param(
        Params._dummy(),
        "binValues",
        "Float values to compare to input column",
        typeConverter=TypeConverters.toListFloat,
    )

    binLabels = Param(
        Params._dummy(),
        "binLabels",
        "Bin labels to use when binning",
        typeConverter=TypeConverters.toList,
    )

    defaultLabel = Param(
        Params._dummy(),
        "defaultLabel",
        "Default label to use when binning",
        typeConverter=TypeConverters.identity,
    )

    def _check_params_size(self, param_name: str, param_value: List[Any]) -> None:
        """
        Checks that the length of the given parameter is the same as the length of
        the other parameters.

        Used to ensure that the parameters are consistent with each other.

        :param param_name: Name of the parameter to check.
        :param param_value: Value of the parameter to check.
        :returns: None
        :raises ValueError: If the length of the given parameter is not the same as
        the length of the other parameters.
        """
        names_to_check = ["conditionOperators", "binValues", "binLabels"]
        names_to_check.remove(param_name)
        for name in names_to_check:
            if self.isDefined(name):
                if len(param_value) != len(self.getOrDefault(name)):
                    raise ValueError(
                        f"""{param_name} must have the same length as {name} but got
                        {len(param_value)} and {len(self.getOrDefault(name))}"""
                    )

    def setConditionOperators(self, value: List[str]) -> "BinParams":
        """
        Sets the conditionOperators parameter.

        :param value: List of string values describing the operator to use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        :returns: Instance of class mixed in.
        """
        allowed_operators = ["eq", "neq", "lt", "gt", "leq", "geq"]
        if any([v not in allowed_operators for v in value]):
            raise ValueError(
                f"""All conditionOperators must be one of {allowed_operators},
                but got {value}"""
            )
        self._check_params_size("conditionOperators", value)
        return self._set(conditionOperators=value)

    def setBinValues(self, value: List[float]) -> "BinParams":
        """
        Sets the binValues parameter.

        :param value: List of float values to compare to input column
        :returns: Instance of class mixed in.
        """
        self._check_params_size("binValues", value)
        return self._set(binValues=value)

    def setBinLabels(self, value: List[Union[float, int, str]]) -> "BinParams":
        """
        Sets the binLabels parameter.

        :param value: List of string values use when binning.
        :returns: Instance of class mixed in.
        """
        self._check_params_size("binLabels", value)
        return self._set(binLabels=value)

    def setDefaultLabel(self, value: Union[float, int, str]) -> "BinParams":
        """
        Sets the defaultLabel parameter.

        :param value: Default label to use when binning.
        :returns: Instance of class mixed in.
        """
        return self._set(defaultLabel=value)

    def getConditionOperators(self) -> List[str]:
        """
        Gets the conditionOperators parameter.

        :returns: List of string values describing the operator to use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        """
        return self.getOrDefault(self.conditionOperators)

    def getBinValues(self) -> List[float]:
        """
        Gets the binValues parameter.

        :returns: List of float values to compare to input column
        """
        return self.getOrDefault(self.binValues)

    def getBinLabels(self) -> List[Union[float, int, str]]:
        """
        Gets the binLabels parameter.

        :returns: List of string values use when binning.
        """
        return self.getOrDefault(self.binLabels)

    def getDefaultLabel(self) -> Union[float, int, str]:
        """
        Gets the defaultLabel parameter.

        :returns: Default label to use when binning.
        """
        return self.getOrDefault(self.defaultLabel)


class BinTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    BinParams,
):
    """
    Bin Spark Transformer for use in Spark pipelines.
    This transformer performs a binning operation on a column in a Spark dataframe.

    The binning operation is performed by comparing the input column to a list of
    values using a list of operators. The bin label corresponding to the first
    condition that evaluates to True is returned.

    If no conditions evaluate to True, the default label is returned.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        conditionOperators: Optional[List[str]] = None,
        binValues: Optional[List[float]] = None,
        binLabels: Optional[List[Union[float, int, str]]] = None,
        defaultLabel: Optional[Union[float, int, str]] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes a BinTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param conditionOperators: List of string values describing the operator to
        use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        :param binValues: Float values to compare to input column.
        :param binLabels: Bin labels to use when binning.
        :param defaultLabel: Default label to use when binning.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
            ByteType(),
            ShortType(),
            IntegerType(),
            LongType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which uses the binValues and binLabels parameters to bin the input column
        according to the conditionOperators parameter.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        condition_operators = [
            get_condition_operator(c) for c in self.getConditionOperators()
        ]
        bin_values = self.getBinValues()
        bin_labels = self.getBinLabels()

        def bin_func(x: Column) -> Column:
            """
            Perfoms the binning of a given column x.
            :param x: Column to bin.
            :returns: Binned column.
            """
            bin_output = F.lit(self.getDefaultLabel())
            # Loop through the conditions.
            # Reverse the list of conditions so that we start from the last condition
            # and work backwards. This ensures that the first condition that is met
            # is the one that is used.
            conds = zip(condition_operators[::-1], bin_values[::-1], bin_labels[::-1])

            for cond_op, value, label in conds:
                bin_output = F.when(cond_op(x, value), F.lit(label)).otherwise(
                    bin_output,
                )
            return bin_output

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: bin_func(x),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the bin transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs the binning operation.
        """
        return BinLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            condition_operators=self.getConditionOperators(),
            bin_values=self.getBinValues(),
            bin_labels=self.getBinLabels(),
            default_label=self.getDefaultLabel(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class HaversineDistanceLayer(BaseLayer):
    """
    Computes the haversine distance operation on a given input tensor.
    If lat_lon_constant is not set, inputs must be a list of 4 tensors,
    in the order of lat1, lon1, lat2, lon2.
    If lat_lon_constant is set, inputs must be a tensor of 2 tensors,
    in the order of lat1, lon1.

    We DO NOT check if the lat/lon values are out of bounds.
    For lat, this is [-90, 90] and for lon, this is [-180, 180].
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        lat_lon_constant: List[float] = None,
        unit: str = "km",
        **kwargs,
    ) -> None:
        """
        Initializes the HaversineDistanceLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param lat_lon_constant: The lat/lons to use in the haversine distance.
        :param unit: The unit of the distance. Must be either 'km' or 'miles'.
        calculation. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if lat_lon_constant is not None and len(lat_lon_constant) != 2:
            raise ValueError("If set, lat_lon_constant must be a list of 2 floats")
        self.lat_lon_constant = lat_lon_constant
        if unit not in ["km", "miles"]:
            raise ValueError("unit must be either 'km' or 'miles'")
        self.unit = unit
        self.earth_radius = 6371.0 if unit == "km" else 3958.8

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bfloat16, tf.float16, tf.float32, tf.float64]

    @staticmethod
    def get_radians(degrees: Tensor) -> Tensor:
        """
        Converts degrees tensor to radians. We need to cast to float64 otherwise
        pi / 180 will lose precision.

        :param degrees: Tensor of degrees.
        :returns: Tensor of radians.
        """
        return tf.cast(degrees, dtype=tf.float64) * tf.constant(
            math.pi / 180, dtype=tf.float64
        )

    def compute_haversine_distance(
        self, lat1: Tensor, lon1: Tensor, lat2: Tensor, lon2: Tensor
    ) -> Tensor:
        """
        Computes the haversine distance between two lat/lon pairs.

        :param lat1: Tensor of latitudes of the first point.
        :param lon1: Tensor of longitudes of the first point.
        :param lat2: Tensor of latitudes of the second point.
        :param lon2: Tensor of longitudes of the second point.
        :returns: Tensor of haversine distances.
        """
        lat1_radians = self.get_radians(lat1)
        lon1_radians = self.get_radians(lon1)
        lat2_radians = self.get_radians(lat2)
        lon2_radians = self.get_radians(lon2)

        lat_diff = lat2_radians - lat1_radians
        lon_diff = lon2_radians - lon1_radians

        a = tf.math.pow(tf.math.sin(lat_diff / 2.0), 2.0) + tf.math.cos(
            lat1_radians
        ) * tf.math.cos(lat2_radians) * tf.math.pow(tf.math.sin(lon_diff / 2.0), 2.0)
        c = 2.0 * tf.math.asin(pow(a, 0.5))
        # Radius of earth in kilometers.
        r = tf.constant(self.earth_radius, dtype=c.dtype)
        return c * r

    @enforce_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Computes the haversine distance between two lat/lon pairs.

        Decorated with @enforce_multiple_tensor_input to ensure that the input
        is an iterable of tensors. Raises an error if a single tensor is passed.

        After decoration, we check the length of the inputs to ensure we have the right
        number of lat/lon tensors.

        :param inputs: Iterable of tensors.
        :returns: Tensor of haversine distances.
        """
        if self.lat_lon_constant is not None:
            if not isinstance(inputs, list) or len(inputs) != 2:
                raise ValueError(
                    """If lat_lon_constant is set,
                inputs must be a list of 2 tensors"""
                )
            return self.compute_haversine_distance(
                inputs[0],
                inputs[1],
                tf.constant(self.lat_lon_constant[0]),
                tf.constant(self.lat_lon_constant[1]),
            )
        else:
            if not isinstance(inputs, list) or len(inputs) != 4:
                raise ValueError(
                    """If lat_lon_constant is not set,
                inputs must be a list of 4 tensors"""
                )
            return self.compute_haversine_distance(
                inputs[0],
                inputs[1],
                inputs[2],
                inputs[3],
            )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the HaversineDistance layer.
        Used for saving and loading from a model.

        Specifically, we add the `lat_lon_constant` and `unit` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"lat_lon_constant": self.lat_lon_constant, "unit": self.unit})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_multiple_tensor_input, reshape_to_equal_rank

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(kamae.__name__)
class ArrayConcatenateLayer(BaseLayer):
    """
    Performs a concatenation of the input tensors.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        axis: int = -1,
        auto_broadcast: bool = False,
        **kwargs,
    ) -> None:
        """
        Initialises the ArrayConcatenateLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: Axis to concatenate on. Defaults to -1.
        :param auto_broadcast: If `True`, will broadcast the input tensors to the
        biggest rank before concatenating. Defaults to `False`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if auto_broadcast and axis != -1:
            raise ValueError("auto_broadcast is only supported for axis=-1")
        self.axis = axis
        self.auto_broadcast = auto_broadcast

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer. Returns `None` as the
        compatible dtypes are not restricted.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Concatenates the input tensors along the specified axis.
        If auto_broadcast is set to True, the tensors are broadcasted to the
        same rank before concatenating.

        Decorated with `@enforce_multiple_tensor_input` to ensure that the input
        is an iterable of tensors. Raises an error if a single tensor is passed
        in.

        :param inputs: Iterable of tensors to concatenate.
        :returns: Concatenated tensor.
        """
        if self.auto_broadcast:
            # Determine the maximum rank statically
            max_rank = max([len(tensor.shape) for tensor in inputs])

            # Reshape all tensors to the same rank, so to calculate later the max_shape
            # WARNING: It assumes that order of inputs and reshaped_inputs is the same!
            reshaped_inputs = reshape_to_equal_rank(inputs)

            # Check the maximum static shape (i.e. with None being the biggest number)
            # except the last one to concat. Here we use the static tensor.shape.
            max_static_shape = []
            for i in range(max_rank - 1):
                shapes = [x.shape[i] for x in reshaped_inputs]
                if None in shapes:
                    max_static_shape.append(None)
                else:
                    max_static_shape.append(max(shapes))

            # Determine the maximum dynamic shape for each dimension, except last one
            # Since shapes can be dynamic (None), we need to use tf.shape
            max_dynamic_shape = []
            for i in range(max_rank - 1):
                shapes = [tf.shape(x)[i] for x in reshaped_inputs]
                max_dynamic_shape.append(tf.reduce_max(shapes))

            # Broadcast tensors to the maximum dynamic shape if the static is different
            # WARNING: It assumes that when the static shapes of two tensors are None
            # at a given rank, the dynamic shapes are the same.
            for idx, x in enumerate(reshaped_inputs):
                x_static_shape = x.shape[:-1]
                if x_static_shape != max_static_shape:
                    last_dim = x.shape[-1]
                    broadcast_shape = tf.concat([max_dynamic_shape, [last_dim]], axis=0)
                    broadcasted_x = tf.broadcast_to(x, broadcast_shape)
                    reshaped_inputs[idx] = broadcasted_x
            inputs = reshaped_inputs

        return tf.concat(inputs, axis=self.axis)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the VectorConcat layer.
        Used for saving and loading from a model.

        Specifically, adds the `axis` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "axis": self.axis,
                "auto_broadcast": self.auto_broadcast,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType

from kamae.spark.params import (
    MultiInputSingleOutputParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import ExponentLayer

from .base import BaseTransformer


class ExponentParams(Params):
    """
    Mixin class containing alpha parameter needed for exponent transform layers.
    """

    exponent = Param(
        Params._dummy(),
        "exponent",
        "Value to use in exponent transform: x^exponent",
        typeConverter=TypeConverters.toFloat,
    )

    def setExponent(self, value: float) -> "ExponentParams":
        """
        Sets the exponent parameter.

        :param value: Float value to use in exponent transform: x^exponent.
        :returns: Instance of class mixed in.
        """
        return self._set(exponent=value)

    def getExponent(self) -> float:
        """
        Gets the exponent parameter.

        :returns: Float value of exponent used in exponent transform.
        """
        return self.getOrDefault(self.exponent)


class ExponentTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    ExponentParams,
):
    """
    Exponent Spark Transformer for use in Spark pipelines.
    This transformer applies x^exponent in the case of single input and or x^y in the
    case of two inputs.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        exponent: Optional[float] = None,
    ) -> None:
        """
        Initializes an ExponentTransformer transformer.

        :param inputCol: Input column name. Only used if inputCols is not specified.
        If specified, we raise this column by the exponent.
        :param inputCols: Input column names. If provided, we raise the first column by
        the second column. Must have exactly two columns.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param exponent: Optional exponent/power to raise the input to. If not provided,
        then two input columns are required.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(exponent=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def setInputCols(self, value: List[str]) -> "ExponentTransformer":
        """
        Overrides setting the input columns for the transformer.
        Throws an error if we do not have exactly two input columns.

        :param value: List of input columns.
        :returns: Class instance.
        """
        if len(value) != 2:
            raise ValueError(
                """When setting inputCols for ExponentTransformer,
                there must be exactly two input columns."""
            )
        return self._set(inputCols=value)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which applies the exponent transform to the input column(s).

        If one column is provided via inputCol, we raise that column to the power of
        the exponent parameter. If two columns are provided via inputCols,
        we raise the first column to the power of the second column.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self.get_multiple_input_cols(
            constant_param_name="exponent", input_cols_limit=2
        )
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: F.pow(x[input_col_names[0]], x[input_col_names[1]]),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the exp value transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs an exp value operation.
        """
        return ExponentLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            exponent=self.getExponent(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .array_utils import (  # noqa: F401
    broadcast_scalar_column_to_array,
    broadcast_scalar_column_to_array_with_inner_singleton_array,
    build_udf_return_type,
    flatten_nested_arrays,
    get_array_nesting_level,
    get_array_nesting_level_and_element_dtype,
    get_element_type,
    nested_arrays_zip,
    nested_lambda,
    nested_transform,
)
from .indexer_utils import (  # noqa: F401
    collect_labels_array,
    collect_labels_array_from_multiple_columns,
)
from .list_utils import (  # noqa: F401
    check_listwise_columns,
    get_listwise_condition_and_window,
)
from .transform_utils import (  # noqa: F401
    multi_input_single_output_array_transform,
    multi_input_single_output_scalar_transform,
    single_input_single_output_array_transform,
    single_input_single_output_array_udf_transform,
    single_input_single_output_scalar_transform,
    single_input_single_output_scalar_udf_transform,
)
from .user_defined_functions import (  # noqa: F401
    hash_udf,
    indexer_udf,
    one_hot_encoding_udf,
    ordinal_array_encode_udf,
)

from .scaler_utils import (  # noqa: F401 # isort:skip
    construct_nested_elements_for_scaling,
)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.tensorflow.layers import IdentityLayer

from .base import BaseTransformer


class IdentityTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
):
    """
    IdentityLayer Spark Transformer for use in Spark pipelines.
    This transformer simply passes the input to the output unchanged.
    Used for cases where you want to keep the input the same.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes an IdentityTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the same as the column with name `inputCol`.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        return dataset.withColumn(self.getOutputCol(), F.col(self.getInputCol()))

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the identity transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs an IdentityLayer operation.
        """
        # Tensorflow <= 2.11 does not contain tf.keras.layers.IdentityLayer
        # so we use a lambda layer instead.
        # When we have a subclassed identity layer, we can use that.
        return IdentityLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import numpy as np
import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import MaskValueParams, SingleInputSingleOutputParams
from kamae.spark.transformers import StandardScaleTransformer
from kamae.spark.utils import construct_nested_elements_for_scaling

from .base import BaseEstimator


class StandardScaleEstimator(
    BaseEstimator,
    SingleInputSingleOutputParams,
    MaskValueParams,
):
    """
    Standard scaler estimator for use in Spark pipelines.
    This estimator is used to calculate the mean and standard deviation of the input
    feature column. When fit is called it returns a StandardScaleTransformer
    which can be used to standardize/transform additional features.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        maskValue: Optional[float] = None,
    ) -> None:
        """
        Initializes a StandardScaleEstimator estimator.
        Sets all parameters to given inputs.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
         in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(maskValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _fit(self, dataset) -> "StandardScaleTransformer":
        """
        Fits the StandardScaleEstimator estimator to the given dataset.
        Calculates the mean and standard deviation of the input feature column and
        returns a StandardScaleTransformer with the mean and standard deviation set.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: StandardScaleTransformer instance with mean & standard deviation set.
        """
        input_column_type = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(input_column_type, ArrayType):
            input_col = F.array(F.col(self.getInputCol()))
            input_column_type = ArrayType(input_column_type)
        else:
            input_col = F.col(self.getInputCol())

        # Collect a single row to driver and get the length.
        # We assume all subsequent rows have the same length.
        array_size = np.array((dataset.select(input_col).first()[0])).shape[-1]

        element_struct = construct_nested_elements_for_scaling(
            column=input_col,
            column_datatype=input_column_type,
            array_dim=array_size,
        )

        mean_cols = [
            F.mean(
                F.when(
                    F.col(f"element_struct.element_{i}") == F.lit(self.getMaskValue()),
                    F.lit(None),
                ).otherwise(F.col(f"element_struct.element_{i}"))
            ).alias(f"mean_{i}")
            for i in range(1, array_size + 1)
        ]

        stddev_cols = [
            F.stddev_pop(
                F.when(
                    F.col(f"element_struct.element_{i}") == F.lit(self.getMaskValue()),
                    F.lit(None),
                ).otherwise(F.col(f"element_struct.element_{i}"))
            ).alias(f"stddev_{i}")
            for i in range(1, array_size + 1)
        ]

        metric_cols = mean_cols + stddev_cols

        mean_and_stddev_dict = (
            dataset.select(element_struct).agg(*metric_cols).first().asDict()
        )
        mean = [mean_and_stddev_dict[f"mean_{i}"] for i in range(1, array_size + 1)]
        stddev = [mean_and_stddev_dict[f"stddev_{i}"] for i in range(1, array_size + 1)]

        return StandardScaleTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            mean=mean,
            stddev=stddev,
            maskValue=self.getMaskValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class OneHotEncodeLayer(BaseLayer):
    """
    Performs a one-hot encoding of a string input tensor.

    Encodes each individual element in the input into an
    array the same size as the vocabulary, containing a 1 at the element
    index. If the last dimension is size 1, will encode on that
    dimension. If the last dimension is not size 1, will append a new
    dimension for the encoded output.
    """

    def __init__(
        self,
        vocabulary: Union[str, List[str]],
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        mask_token: str = None,
        num_oov_indices: int = 1,
        drop_unseen: bool = False,
        encoding: str = "utf-8",
        **kwargs,
    ) -> None:
        """
        Intialises the OneHotLayer layer.

        :param vocabulary: Either an array of strings or a string path to a
        text file. If passing an array, can pass a tuple, list, 1D numpy array,
        or 1D tensor containing the string vocbulary terms. If passing a file
        path, the file should contain one line per term in the vocabulary.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param mask_token: A token that represents masked inputs. The token is included
        in vocabulary and mapped to index 0. If set to None, no mask term will be added.
        Defaults to `None`.
        :param num_oov_indices: The number of out-of-vocabulary indices to use. The
        out-of-vocabulary indices are used to represent unseen labels and are placed at
        the beginning of the one-hot encoding. Defaults to 1.
        :param drop_unseen: Whether to drop unseen label indices. If set to True, the
        layer will not add an extra dimension for unseen labels in the one-hot
        encoding. Defaults to False.
        :param encoding: The text encoding to use to interpret the input strings.
        Defaults to `"utf-8"`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.num_oov_indices = num_oov_indices
        self.vocabulary = vocabulary
        self.drop_unseen = drop_unseen
        self.mask_token = mask_token
        self.encoding = encoding
        self.lookup_layer = tf.keras.layers.StringLookup(
            vocabulary=self.vocabulary,
            output_mode="int",
            num_oov_indices=self.num_oov_indices,
            mask_token=self.mask_token,
            encoding=self.encoding,
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.int16, tf.int32, tf.int64, tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the one-hot encoding on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to one-hot encode.
        :returns: One-hot encoded input tensor.
        """
        casted_inputs = (
            tf.strings.as_string(inputs, scientific=False)
            if inputs.dtype != tf.string
            else inputs
        )
        indexed_inputs = self.lookup_layer(casted_inputs)
        mask_offset = 1 if self.mask_token is not None else 0

        # If last dimension to encode is 1,
        # remove it after one-hot encoding.
        # E.g. (None, None, 1) -> (None, None, 1, N) -> (None, None, N)
        # But (None, None, M) -> (None, None, M, N)
        ohe_depth = len(self.vocabulary) + self.num_oov_indices + mask_offset
        encoded_inputs = (
            tf.squeeze(tf.one_hot(indexed_inputs, ohe_depth), axis=-2)
            if indexed_inputs.get_shape()[-1] == 1
            else tf.one_hot(indexed_inputs, ohe_depth)
        )

        # If drop unseen, slice off the first num_oov_indices + mask_offset columns
        if self.drop_unseen:
            encoded_inputs = encoded_inputs[..., (self.num_oov_indices + mask_offset) :]

        return encoded_inputs

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the OneHot layer.
        Used for saving and loading from a model.

        Specifically adds the `vocabulary`, `num_oov_indices`, `mask_token`, and
        `encoding` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "vocabulary": self.vocabulary,
                "num_oov_indices": self.num_oov_indices,
                "drop_unseen": self.drop_unseen,
                "mask_token": self.mask_token,
                "encoding": self.encoding,
            }
        )
        return config


# TODO: Remove this alias in next breaking change,
#  it is maintained for backwards compatibility
class OneHotLayer(OneHotEncodeLayer):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "OneHotLayer is deprecated and will be removed in a future release. "
            "Use OneHotEncodeLayer instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import (
    ByteType,
    DataType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
)

from kamae.spark.params import (
    MultiInputSingleOutputParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import (
    get_element_type,
    multi_input_single_output_scalar_transform,
)
from kamae.tensorflow.layers import DateAddLayer


class DateAdditionParams(Params):
    """
    Mixin class for a date addition transformer.
    """

    numDays = Param(
        Params._dummy(),
        "numDays",
        "Number of days to add/subtract. Negative values subtract.",
        typeConverter=TypeConverters.toInt,
    )

    def getNumDays(self) -> int:
        """
        Gets the value of the numDays parameter.

        :returns: Number of days to add/subtract.
        """
        return self.getOrDefault(self.numDays)

    def setNumDays(self, value: int) -> "DateAdditionParams":
        """
        Sets the value of the numDays parameter.

        :param value: Number of days to add/subtract.
        :returns: Class instance.
        """
        if self.isDefined("inputCols"):
            raise ValueError("Cannot set numDays if using multiple inputCols.")
        return self._set(numDays=value)


class DateAddTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    DateAdditionParams,
):
    """
    Transformer to add or subtract a static or dynamic (column) number of days
    from a date column.

    WARNING: This transform destroys the time component of the date column.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        numDays: Optional[int] = None,
    ) -> None:
        """
        Initialises the date add transform layer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Layer name. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param numDays: Number of days to add/subtract. Negative values subtract.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(numDays=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def setInputCols(self, value: List[str]) -> "DateAddTransformer":
        """
        Sets the value of the inputCols parameter.

        :param value: Input column names.
        :returns: Class instance.
        """
        if len(value) != 2:
            raise ValueError("If using multiple inputs, exactly two are required.")
        if self.getNumDays() is not None:
            raise ValueError("Cannot use multiple inputs if numDays is set.")
        if self.getInputDtype() is not None:
            raise ValueError(
                """Input auto-casting is set via inputDtype, however multiple inputs are
                being used. Auto-casting inputs is not supported for multiple inputs in
                the DateAddTransformer because the two inputs must be
                different types."""
            )
        return self._set(inputCols=value)

    def setInputDtype(self, value: str) -> "DateAddTransformer":
        """
        Overrides setting the parameter inputDtype to the given string value.

        If multiple input columns are being used, the inputDtype parameter cannot be
        set.

        :param value: String to set the inputDtype parameter to.
        :raises ValueError: If inputCols is set.
        :returns: Instance of class mixed in.
        """
        if self.isDefined("inputCols"):
            raise ValueError(
                """Input auto-casting is not supported for multiple inputs in the
                DateAddTransformer because the two inputs must be different types."""
            )
        return self._set(inputDtype=value)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType(), ByteType(), ShortType(), IntegerType(), LongType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Adds or subtracts a number of days from a date column.

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """
        input_cols = self.get_multiple_input_cols(
            constant_param_name="numDays", input_cols_limit=2
        )
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]
        if not isinstance(get_element_type(input_col_datatypes[0]), StringType):
            raise ValueError(
                f"""Expected input column {input_col_names[0]} to have element type
                StringType, but got {input_col_datatypes[0]}."""
            )
        if not isinstance(
            get_element_type(input_col_datatypes[1]),
            (ByteType, ShortType, IntegerType, LongType),
        ):
            raise ValueError(
                f"""Expected input column {input_col_names[1]} to have element type
                 ByteType, ShortType or IntegerType, but got {input_col_datatypes[1]}.
                 """
            )
        date_col_name = input_col_names[0]
        num_days_col_name = input_col_names[1]
        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            # Cast to int since LongType not supported in date_add, but we want to allow
            # it as an input type.
            func=lambda x: F.date_add(
                x[date_col_name], x[num_days_col_name].cast("int")
            ).cast("string"),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer.

        :returns: DateAddLayer Tensorflow layer.
        """
        return DateAddLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            num_days=self.getNumDays(),
        )
